(*  
    Author:      René Thiemann 
                 Akihisa Yamada
    License:     BSD
*)
section \<open>Checking Polynomial Growth of Matrices\<close>

text \<open>We combine the developed algorithms into a single check-function which
  tries to ensure polynomial growth of a matrix w.r.t. to a given degree.\<close>
 
theory Jordan_Normal_Form_Complexity_Approximation
imports
  "$AFP/Jordan_Normal_Form/Jordan_Normal_Form_Uniqueness"
  "$AFP/Jordan_Normal_Form/Jordan_Normal_Form_Existence"
  "$AFP/Jordan_Normal_Form/Complexity_Carrier"
  "$AFP/Jordan_Normal_Form/Ring_Hom_Matrix"
  "$AFP/Jordan_Normal_Form/Show_Matrix"
  "$AFP/Algebraic_Numbers/Complex_Algebraic_Numbers"
  "$AFP/Polynomial_Factorization/Select_Berlekamp_Hensel_Factorization"
  "$AFP/Algebraic_Numbers/Show_Real_Precise"
  "$AFP/Show/Show_Complex"
  "$AFP/Show/Show_Poly"
  "$AFP/Certification_Monads/Check_Monad"
begin

hide_const (open) Real_Algebraic_Numbers.real_of 

definition analyse_evs :: "nat \<Rightarrow> complex mat \<Rightarrow> complex list \<Rightarrow> shows check" where
  "analyse_evs d A evs \<equiv> check_allm (\<lambda> e. do {
       check (is_norm_le_1 e) (shows ''detected eigenvalue with norm > 1: '' o shows e);
       if is_norm_1 e then 
         check_allm (\<lambda> (n,_). check (n \<le> d) (shows ''detected Jordan-Block of size '' o shows n
           o shows '' for eigenvalue '' o shows e)) (compute_set_of_jordan_blocks A e)
       else succeed
     }) evs"


definition approximate_jordan_blocks_complex_mat :: "nat \<Rightarrow> complex mat \<Rightarrow> shows check" where
  "approximate_jordan_blocks_complex_mat d A \<equiv> let
     n = dim\<^sub>r A;
     cp = char_poly A;
     evso = roots_of_complex_poly cp
     in do {
       check (evso \<noteq> None) (shows ''could not determine Eigenvalues'');
       analyse_evs d A (the evso)
     } <+? (\<lambda> s. shows ''problem in estimating growth rate of matrix\<newline>'' o shows A o 
       shows ''\<newline>with characteristic polynomial '' o shows cp o shows_nl o s)"

lemma approximate_jordan_blocks_complex_mat: assumes A: "A \<in> carrier\<^sub>m n n"
  and res: "isOK(approximate_jordan_blocks_complex_mat d A)"
shows "\<exists>c1 c2. \<forall>k. norm_bound (A ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))" 
proof -
  from char_poly_factorized[OF A] obtain n_as' where "char_poly A = (\<Prod>a\<leftarrow>n_as'. [:- a, 1:])" by auto
  from jordan_nf_exists[OF A this] obtain n_as where jnf: "jordan_nf A n_as" by auto
  from jordan_nf_char_poly[OF this] have cp: "char_poly A = (\<Prod>(n, a)\<leftarrow>n_as. [:- a, 1:] ^ n)" .
  from res
  obtain evs  where roots: "roots_of_complex_poly (char_poly A) = Some evs" 
    and ana_evs: "isOK(analyse_evs d A evs)" 
    unfolding approximate_jordan_blocks_complex_mat_def Let_def by (auto split: prod.splits)
  let ?cp = "char_poly A"
  from degree_monic_char_poly[OF A] have "?cp \<noteq> 0" by auto
  from roots_of_complex_poly[OF roots]
  have rts: "{x. poly ?cp x = 0} = set evs" by auto
  show ?thesis
  proof (rule jordan_nf_matrix_poly_bound[OF A _ _ jnf])
    fix n a
    assume na: "(n,a) \<in> set n_as"
    then obtain as bs where nas: "n_as = as @ (n,a) # bs" by (meson split_list)
    obtain q where cp: "char_poly A = ([:-a,1:]^n) * q" unfolding cp nas by auto
    { 
      assume n: "0 < n"
      with cp na have rt: "poly ?cp a = 0" by auto
      with rts have a: "a \<in> set evs" by auto
      let ?ns = "[n . ev <- evs, norm ev = 1, n <- map fst (compute_set_of_jordan_blocks A ev)]"
      from ana_evs[unfolded analyse_evs_def] a have 
        cb: "\<And> n b . norm a = 1 \<Longrightarrow> (n,b) \<in> set (compute_set_of_jordan_blocks A a) \<Longrightarrow> n \<le> d" and 
        norm_a: "norm a \<le> 1" 
        by auto
      {
        assume n1: "norm a = 1"
        have "n \<le> d"
          by (rule cb[OF n1], unfold compute_set_of_jordan_blocks[OF jnf], insert n na a, auto)
      }
      with norm_a have "norm a \<le> 1" "norm a = 1 \<Longrightarrow> n \<le> d" by auto
    } note * = this
    show "0 < n \<Longrightarrow> cmod a \<le> 1" using * by auto
    show "cmod a = 1 \<Longrightarrow> n \<le> d" using * by (cases "n = 0", auto)
  qed
qed

definition count_ones_check :: "real mat \<Rightarrow> nat \<Rightarrow> bool" where
  "count_ones_check A d \<equiv> upper_triangular A \<and> (let diag = mat_diag A
     in (\<forall> a \<in> set diag. abs a \<le> 1 \<and> (abs a = 1 \<longrightarrow> length (filter (op = a) diag) \<le> d)))"

text \<open>We first try a cheap test whether we have the simple case: an upper-triangular matrix. 
  If the counting-ones criterion succeeds, then we are done, and only otherwise we start the
  full machinery with root-finding algorithms, Jordan-block computation, etc.\<close>

definition approximate_jordan_blocks_real_mat :: "complex list \<Rightarrow> complex poly list \<Rightarrow> nat \<Rightarrow> real mat \<Rightarrow> shows check" where
  "approximate_jordan_blocks_real_mat csamples cpsamples d A \<equiv>
    if count_ones_check A d then succeed else 
     approximate_jordan_blocks_complex_mat d (map\<^sub>m of_real A)"

lemma approximate_jordan_blocks_real_mat: assumes A: "A \<in> carrier\<^sub>m n n"
  and res: "isOK(approximate_jordan_blocks_real_mat csamples cpsamples d A)"
shows "\<exists>c1 c2. \<forall>k. norm_bound (A ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))" 
proof (cases "count_ones_check A d")
  case False
  let ?B = "map\<^sub>m complex_of_real A"
  from res False have res: "isOK(approximate_jordan_blocks_complex_mat d ?B)"
    unfolding approximate_jordan_blocks_real_mat_def by auto
  from A have B: "?B \<in> carrier\<^sub>m n n" by auto
  from approximate_jordan_blocks_complex_mat[OF B res]
  obtain c1 c2 where bnd: "\<And> k. norm_bound (?B ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))" by blast
  show ?thesis
  proof (rule exI[of _ c1], rule exI[of _ c2], intro allI)
    fix k
    let ?c = "(c1 + c2 * real_of_nat k ^ (d - 1))"
    def c \<equiv> ?c
    interpret ring_hom complex_of_real
      by (unfold_locales, auto)
    from bnd[of k] have "norm_bound (?B ^\<^sub>m k) c" unfolding c_def .
    hence "norm_bound (A ^\<^sub>m k) c" unfolding norm_bound_def
      unfolding mat_hom_pow[OF A, symmetric] by auto
    thus "norm_bound (A ^\<^sub>m k) ?c" unfolding c_def .
  qed
next
  case True
  show ?thesis
    by (rule counting_ones_complexity[OF A], insert True[unfolded count_ones_check_def Let_def], auto)
qed

definition mat_estimate_complexity_jb :: "nat \<Rightarrow> 'a :: large_real_ordered_semiring_1 mat \<Rightarrow> shows check" where
  "mat_estimate_complexity_jb d A \<equiv> let B = mat\<^sub>\<real> A in do {
     check (dim\<^sub>r A = dim\<^sub>c A) (shows ''expected square matrix but got \<newline>'' o shows B);
     approximate_jordan_blocks_real_mat [] [] d B
   }"

lemma mat_estimate_complexity_jb_norm_bound: assumes *: "isOK(mat_estimate_complexity_jb d A)"
  shows "\<exists> c1 c2. \<forall> k. norm_bound (mat\<^sub>\<real> A ^\<^sub>m k) (c1 + c2 * of_nat k ^ (d - 1))"
proof -
  note * = *[unfolded mat_estimate_complexity_jb_def Let_def]
  from * obtain n where A: "A \<in> carrier\<^sub>m n n" unfolding mat_carrier_def by auto 
  from * have *: "isOK(approximate_jordan_blocks_real_mat [] [] d (mat\<^sub>\<real> A))" by auto
  let ?B = "mat\<^sub>\<real> A"
  from A have "?B \<in> carrier\<^sub>m n n" by auto
  from approximate_jordan_blocks_real_mat[OF this *] show ?thesis .
qed

context
  fixes A B C :: "'a :: large_real_ordered_semiring_1 mat" and n d e :: nat
  assumes *: "isOK(mat_estimate_complexity_jb e A)"
  and A: "A \<in> carrier\<^sub>m n n" and B: "B \<in> carrier\<^sub>m n n" and C: "C \<in> carrier\<^sub>m n n"
  and d: "d = e - 1"
begin

lemma mat_estimate_complexity_jb_norm_bound_prod: 
  "\<exists> c. \<forall> k. k > 0 \<longrightarrow> norm_bound (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C)) (c * of_nat k ^ d)"
proof -
  let ?R = "mat\<^sub>\<real>"
  let ?A = "?R A"
  let ?B = "?R B"
  let ?C = "?R C"
  from mat_estimate_complexity_jb_norm_bound[OF *, folded d]
  obtain c1 c2 where bnd: "\<And> k. norm_bound (?A ^\<^sub>m k) (c1 + c2 * of_nat k ^ d)" by auto
  def c \<equiv> "(max c1 0) + (max c2 0)"
  from norm_bound_max[of ?B] obtain nb where nb: "norm_bound ?B nb" by auto
  from norm_bound_max[of ?C] obtain nc where nc: "norm_bound ?C nc" by auto
  let ?n = "of_nat n :: real"
  def b \<equiv> "nb * nc * ?n * ?n * c"
  {
    fix k
    assume "k > (0 :: nat)"
    let ?k = "of_nat k :: real"
    from bnd[of k] have "norm_bound (?A ^\<^sub>m k) (c1 + c2 * ?k ^ d)" .
    have "c1 \<le> (max c1 0) * 1 ^ d" by auto
    also have "\<dots> \<le> (max c1 0) * ?k ^ d"
      by (rule mult_left_mono, insert `k > 0`, auto)
    also have "\<dots> + c2 * ?k ^ d \<le> c * ?k ^ d"
      unfolding c_def by (auto simp: field_simps intro: mult_right_mono)
    finally have "c1 + c2 * of_nat k ^ d \<le> c * of_nat k ^ d" by auto
    with bnd[of k] have "norm_bound (?A ^\<^sub>m k) (c * of_nat k ^ d)"
      unfolding norm_bound_def by force
    from norm_bound_mult[OF mat_pow_closed _ this nc]
    have "norm_bound (?A ^\<^sub>m k \<otimes>\<^sub>m ?C) (c * ?k ^ d * nc * ?n)" using A C by auto
    from norm_bound_mult[OF _ mat_mult_mat_closed[OF mat_pow_closed] nb this]
    have "norm_bound (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C)) (b * ?k ^ d)" unfolding b_def using A B C
      by (auto simp: ac_simps)
  } note bnd = this
  thus ?thesis by auto
qed

text \<open>This is the main result for real valued matrices.\<close>

lemma mat_estimate_complexity_jb_norm_mat_sum_prod: 
  "\<exists> c. \<forall> k. k > 0 \<longrightarrow> norm (mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C))) \<le> (c * of_nat k ^ d)"
proof -
  let ?R = "mat\<^sub>\<real>"
  let ?A = "?R A"
  let ?B = "?R B"
  let ?C = "?R C"
  from mat_estimate_complexity_jb_norm_bound_prod obtain c
  where bnd: "\<And> k. k > 0 \<Longrightarrow> norm_bound (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C)) (c * of_nat k ^ d)" by auto
  let ?n = "of_nat n :: real"
  def b \<equiv> "?n * ?n * c" 
  {
    fix k
    let ?nn = "{0 ..< n}"
    let ?g = "\<lambda> i. c * of_nat k ^ d"
    assume "k > (0 :: nat)"    
    from bnd[OF this] have bnd: "norm_bound (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C)) (c * of_nat k ^ d)" .
    have "norm (mat_sum (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C))) \<le> setsum ?g (?nn \<times> ?nn)"
      unfolding mat_sum_def
      by (rule order_trans[OF setsum_norm_le[of _ _ ?g]],
      insert bnd A B C, auto simp: norm_bound_def)
    also have "\<dots> = b * of_nat k ^ d" unfolding b_def setsum_constant
      by (simp add: ac_simps of_nat_mult)
    finally have "norm (mat_sum (?B \<otimes>\<^sub>m (?A ^\<^sub>m k \<otimes>\<^sub>m ?C))) \<le> b * of_nat k ^ d" by auto
  }
  thus ?thesis by auto
qed

text \<open>And via conversion, it also holds for matrices over the intended semiring.\<close>

lemma mat_estimate_complexity_jb_mat_sum_prod: 
  "\<exists> c. \<forall> k. k > 0 \<longrightarrow> mat_sum (B \<otimes>\<^sub>m (A ^\<^sub>m k \<otimes>\<^sub>m C)) \<le> (c * of_nat k ^ d)"
proof -  
  from mat_estimate_complexity_jb_norm_mat_sum_prod obtain c where 
    bnd: "\<And> k. k > 0 \<Longrightarrow> norm (mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C))) \<le> (c * of_nat k ^ d)"
    by auto
  note mat_real_conv = 
    real_embedding.hom_mat_sum 
    real_embedding.mat_hom_pow [of _ n]
    real_embedding.mat_hom_mult[of _ n n]
  {
    fix k
    assume "k > (0 :: nat)"
    let ?sum = "mat_sum (B \<otimes>\<^sub>m (A ^\<^sub>m k \<otimes>\<^sub>m C))"
    have Ak: "A ^\<^sub>m k \<in> carrier\<^sub>m n n" using A by simp
    have AkC: "A ^\<^sub>m k \<otimes>\<^sub>m C \<in> carrier\<^sub>m n n" using Ak C by simp
    let ?ck = "c * of_nat k ^ d"
    let ?cck = "of_int \<lceil>c\<rceil> * of_nat k ^ d :: real"
    from bnd[OF `k > 0`] have "mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C)) \<le> ?ck" by auto
    also have "mat_sum (mat\<^sub>\<real> B \<otimes>\<^sub>m (mat\<^sub>\<real> A ^\<^sub>m k \<otimes>\<^sub>m mat\<^sub>\<real> C)) = 
      real_of ?sum" 
      using Ak AkC A B C by (simp add: mat_real_conv)
    finally have "real_of ?sum \<le> ?ck" .
    from real_le[OF this] have le: "?sum \<le> of_int \<lceil>?ck\<rceil>" by auto
    have "?ck \<le> ?cck"
      by (rule mult_right_mono, auto)
    hence "\<lceil>?ck\<rceil> \<le> \<lceil>?cck\<rceil>" by linarith
    hence "of_int \<lceil>?ck\<rceil> \<le> ((of_int \<lceil>?cck\<rceil>) :: 'a)" 
      unfolding of_int_le_iff .
    with le have le: "?sum \<le> of_int \<lceil>?cck\<rceil>" by linarith
    also have "\<dots> = of_int \<lceil>c\<rceil> * of_nat k ^ d"
      by (metis (no_types, hide_lams) ceiling_of_int of_int_mult of_int_of_nat_eq of_nat_power)
    finally have "?sum \<le> of_int \<lceil>c\<rceil> * of_nat k ^ d" .
  }
  thus ?thesis by auto
qed
end

end
